/*
                                  coadd.c

*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*
*       Part of:        Swarp
*
*       Author:         E.BERTIN (IAP)
*
*       Contents:       Coaddition routines
*
*       Last modify:    29/09/2002
*
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

/*
  04.09.2008:
  Inclusion of the co-addition mode SUM - a simple sum of the input frames
*/

#ifdef HAVE_CONFIG_H
#include	"config.h"
#endif

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "define.h"
#include "types.h"
#include "globals.h"
#include "fits/fitscat.h"
#include "fitswcs.h"
#include "coadd.h"
#include "data.h"
#include "field.h"
#include "interpolate.h"
#include "prefs.h"
#include "weight.h"
#include "wcs/wcs.h"

char	padbuf[FBSIZE];

/******* coadd_fields *********************************************************
PROTO	int coadd_fields(fieldstruct **infield, fieldstruct **inwfield,
			int ninput,
			fieldstruct *outfield, fieldstruct *outwfield,
			coaddenum coaddtype, PIXTYPE wthresh)
PURPOSE	Coadd images.
INPUT	Input field ptr array,
	Input weight field ptr array,
	number of input fields,
	Output field ptr,
	Output weight field ptr,
	Coaddition type.
OUTPUT	RETURN_OK if no error, or RETURN_ERROR in case of non-fatal error(s).
NOTES   -.
AUTHOR  E. Bertin (IAP)
VERSION 29/09/2002
 ***/
int coadd_fields(fieldstruct **infield, fieldstruct **inwfield,	int ninput,
			fieldstruct *outfield, fieldstruct *outwfield,
			coaddenum coaddtype, PIXTYPE wthresh)

  {
   fieldstruct		*field;
   tabstruct		*tab;
   wcsstruct		*wcs;
   PIXTYPE		*emptybuf,*emptywbuf, *inbuf,*inwbuf, *outbuf,*outwbuf,
			*pix,*wpix,
			*multibuf,*multiwbuf, *multi,*multiw;
   long			offbeg, offend;
   int			rawmin[NAXIS], rawmax[NAXIS], rawpos[NAXIS],
			rawmaxmax[NAXIS];
   int			d,n, outwidth,multiwidth, width, height,
			naxis,nlines, x,y,yc, size, offset, wflag;

  naxis = outfield->tab->naxis;
/* The output width is the useful length of the NAXIS1 axis */
  width = outfield->width;
/* The output ``height'' is the product of all other axis lengths */
  height = outfield->height;

/*-- Find global extrema of mapped image */
  for (d=0; d<naxis; d++)
    {
    rawmin[d] = 1<<30;
    rawmax[d] = 0;
    for (n = 0; n<ninput; n++)
      {
      if (infield[n]->wcs->outmin[d] < rawmin[d])
        rawmin[d] = (int)infield[n]->wcs->outmin[d];
      if (infield[n]->wcs->outmax[d] > rawmax[d])
        rawmax[d] = (int)infield[n]->wcs->outmax[d];
      }
    rawpos[d] = 1;
    rawmaxmax[d] = outfield->wcs->naxisn[d];
    }

/* Empty pixels at the beginning of each line */
  offbeg = rawmin[0]-1;
  if (offbeg < 0)
    offbeg = 0;
  if (offbeg>outfield->wcs->naxisn[0])
    offbeg = outfield->wcs->naxisn[0];

/* Empty pixels at the end of each line */
  offend = outfield->wcs->naxisn[0] - rawmax[0];
  if (offend < 0)
    offend = 0;
  if (offend>outfield->wcs->naxisn[0])
    offend = outfield->wcs->naxisn[0];

  outwidth = outfield->wcs->naxisn[0] - offend - offbeg;
  nlines = 1;
  if (nlines<1)
    nlines = 1;
  multiwidth = outwidth*ninput;

/* Open output file and save header */
  if (open_cat(outfield->cat, WRITE_ONLY) != RETURN_OK)
      error(EXIT_FAILURE, "*Error*: cannot open for writing ",
		outfield->filename);
  QFWRITE(outfield->tab->headbuf, outfield->tab->headnblock*FBSIZE,
	outfield->cat->file, outfield->filename);

/* Open output weight file and save header */
  outwfield->sigfac = (double)1.0;	/* A possible scaling among others */
  if (open_cat(outwfield->cat, WRITE_ONLY) != RETURN_OK)
    error(EXIT_FAILURE, "*Error*: cannot open for writing ",
		outwfield->filename);
  QFWRITE(outwfield->tab->headbuf, outwfield->tab->headnblock*FBSIZE,
	outwfield->cat->file, outwfield->filename);

/* Only for WRITING the weights */
  set_weightconv(outwfield);

/* Re-open Input files */
  for (n = 0; n<ninput; n++)
    {
    if (open_cat(infield[n]->cat, READ_ONLY) != RETURN_OK)
      error(EXIT_FAILURE, "*Error*: cannot open for reading ",
		infield[n]->filename);
    tab = infield[n]->cat->tab;
    tab->bodypos = tab->headnblock*FBSIZE;
    QFSEEK(tab->cat->file, tab->bodypos, SEEK_SET, infield[n]->filename);
    if (inwfield[n])
      {
      if (open_cat(inwfield[n]->cat, READ_ONLY) != RETURN_OK)
        error(EXIT_FAILURE, "*Error*: cannot open for reading ",
		inwfield[n]->filename);
      tab = inwfield[n]->cat->tab;
      tab->bodypos = tab->headnblock*FBSIZE;
      QFSEEK(tab->cat->file, tab->bodypos, SEEK_SET, inwfield[n]->filename);
      wflag = 1;
      }
    else
      wflag = 0;
    }

  if (outwidth)
    {
/*-- Allocate memory for the output buffers that contain the input data in */
/*-- internal format (PIXTYPE) */
    QMALLOC(inbuf, PIXTYPE, nlines*outwidth);
    if (wflag)
      QMALLOC(inwbuf, PIXTYPE, nlines*outwidth);
/*-- Allocate memory for the output buffers that contain the final data in */
/*-- internal format (PIXTYPE) */
    QMALLOC(outbuf, PIXTYPE, nlines*outwidth);
    QMALLOC(outwbuf, PIXTYPE, nlines*outwidth);
/*-- Allocate memory for the "multi-buffer" storing resampled pixels from */
/*-- all images for the current line(s), prior to co-addition */
    QMALLOC(multibuf, PIXTYPE, nlines*multiwidth);
    QMALLOC(multiwbuf, PIXTYPE, nlines*multiwidth);
    }
/* Allocate memory for the output buffers that contain "empty data" */
  QCALLOC(emptybuf, PIXTYPE, width);
  QCALLOC(emptywbuf, PIXTYPE, width);

  multiw = NULL;

/* Loop over output ``lines'': this can be over more than 1 (Y) dimension */
  for (y = 0; y<height; y++)
    {
    if (!((y+1)%16))
      NPRINTF(OUTPUT, "\33[1M> Co-adding line:%7d / %-7d\n\33[1A",
	y+1, height);
/*---- Skip empty lines */
    for (d=naxis; --d;)
      if (rawpos[d]<rawmin[d] || rawpos[d]>rawmax[d])
        break;
    if (d>0)
      {
      write_body(outfield->tab, emptybuf, (long)width);
      write_body(outwfield->tab, emptywbuf, (long)width);
      }
    else
      {
/*---- Skip empty pixels */
      if (offbeg)
        {
        write_body(outfield->tab, emptybuf, offbeg);
        write_body(outwfield->tab, emptywbuf, offbeg);
        }

/*---- The current "physical" line is y modulo nlines */
      yc = y%nlines;

      if (outwidth)
        {
/*------ Initialize output data line */
        memset(multibuf+yc*multiwidth, 0, multiwidth*sizeof(PIXTYPE));
        multiw = multiwbuf+yc*multiwidth;
        for (x=multiwidth; x--;)
          *(multiw++) = BIG;

/*------ Loop over input images */
        for (n = 0; n<ninput; n++)
          {
          field = infield[n];
          wcs = field->wcs;
/*-------- Check that the present line is contained in the field */
          for (d=naxis; --d;)
            if (rawpos[d]<wcs->outmin[d] || rawpos[d]>wcs->outmax[d])
              break;
          if (d>0)
            continue;
          read_body(field->tab, pix=inbuf, field->width);
          if (wflag)
            read_body(inwfield[n]->tab, wpix=inwbuf, field->width);
          offset = yc*multiwidth+(wcs->outmin[0]-1-offbeg)*ninput+n;
          multi = multibuf+offset;
          multiw = multiwbuf+offset;
          for (x=field->width; x--;)
            {
            *multi = *(pix++);
            multi += ninput;
/*--------- Convert weight to variance */
            if (wflag)
              {
              *multiw = (*wpix> 0.0)? 1.0/(*wpix) : BIG;
              wpix++;
              }
            else
              *multiw = 1.0;
            multiw += ninput;
            }
          }

/*------ Now perform the coaddition itself */
        multi = multibuf+yc*multiwidth;
        multiw = multiwbuf+yc*multiwidth;
        pix = outbuf;
        wpix = outwbuf;
        for (x=outwidth; x--;)
          {
          coadd_pix(multi,multiw, ninput, pix++,wpix++,coaddtype,wthresh);
          multi += ninput;
          multiw += ninput;
          }

        write_body(outfield->tab, outbuf, outwidth);
        var_to_weight(outwbuf, outwidth);
        write_body(outwfield->tab, outwbuf, outwidth);
        }

/*---- Skip empty pixels */
      if (offend)
        {
        write_body(outfield->tab, emptybuf, offend);
        write_body(outwfield->tab, emptywbuf, offend);
        }
      }

/*-- Update coordinate vector */
    for (d=1; d<naxis; d++)
      if ((++rawpos[d])<=rawmaxmax[d])
        break;
      else
        rawpos[d] = 1;
    }

/* FITS padding*/
  size = PADEXTRA(outfield->tab->tabsize);
  if (size)
    QFWRITE(padbuf, (size_t)size, outfield->cat->file, outfield->filename);
  size = PADEXTRA(outwfield->tab->tabsize);
  if (size)
    QFWRITE(padbuf, (size_t)size, outwfield->cat->file, outwfield->filename);

/* Close files */
  close_cat(outfield->cat);
  close_cat(outwfield->cat);
  for (n = 0; n<ninput; n++)
    {
    close_cat(infield[n]->cat);
    if (wflag)
      close_cat(inwfield[n]->cat);
    }

/* Free Buffers */
  free(emptybuf);
  free(emptywbuf);
  if (outwidth)
    {
    free(multibuf);
    free(multiwbuf);
    free(inbuf);
    if (wflag)
      free(inwbuf);
    free(outbuf);
    free(outwbuf);
    }

  return RETURN_OK;
  }


/******* coadd_pix ************************************************************
PROTO	int coadd_pix(PIXTYPE *inpix, PIXTYPE *inwpix, int ninput,
		PIXTYPE *outpix , PIXTYPE *outwpix, coaddenum coaddtype,
		PIXTYPE wthresh)
PURPOSE	Coadd images.
INPUT	Input pixel array ptr,
	Input weight ptr array,
	number of input fields,
	Output pixel ptr,
	Output weight ptr,
	Coaddition type.
OUTPUT	RETURN_OK if no error, or RETURN_ERROR in case of non-fatal error(s).
NOTES   -.
AUTHOR  E. Bertin (IAP)
VERSION 01/08/2001
 ***/
int coadd_pix(PIXTYPE *inpix, PIXTYPE *inwpix, int ninput,
		PIXTYPE *outpix , PIXTYPE *outwpix, coaddenum coaddtype,
		PIXTYPE wthresh)

  {
   static PIXTYPE	pixstack[MAXINFIELD];
   PIXTYPE		*pixt;
   double		val, val2, wval,wval2,wval3;
   int			i, ninput2;

  switch(coaddtype)
    {
    case COADD_MEDIAN:
      ninput2 = 0;
      wval = wval3 = 0.0;
      pixt = pixstack;
      for (i=ninput; i--;)
        {
        val2 = *(inpix++);
        wval2 = *(inwpix++);        
        if (wval2 < wthresh)
          {
          *(pixt++) = val2;
          wval += 1.0/sqrt(wval2);
          wval3 += wval2;
          ninput2++;
          }      
        }
      if (ninput2)
        {
        *outpix = hmedian(pixstack, ninput2);
/*------ We assume Gaussian input noise */
        *outwpix = ninput2>2?
		(PI*ninput2*ninput2/(2*wval*wval*(ninput2+((ninput2&1)?(PI/2-1)
			:(PI-2)))))
	  : wval3/(ninput2*ninput2);
        }
      else
        {
        *outpix = 0.0;
        *outwpix = BIG;
        }
      break;
    case COADD_AVERAGE:
      ninput2 = 0;
      wval = val = 0.0;
      for (i=ninput; i--;)
        {
        val2 = *(inpix++);
        wval2 = *(inwpix++);
        if (wval2<wthresh)
          {
          ninput2++;
          val += val2;
          wval += wval2;
          }
        }
      if (ninput2)
        {
        *outpix = val/ninput2;
        *outwpix = wval/(ninput2*ninput2);
        }
      else
        {
        *outpix = 0.0;
        *outwpix = BIG;
        }
      break;
    case COADD_MIN:
      ninput2 = 0;
      val = BIG;
      wval = 0.0;
      for (i=ninput; i--;)
        {
        val2 = *(inpix++);
        wval2 = *(inwpix++);
        if (wval2<wthresh && val2<val)
          {
          ninput2++;
          val = val2;
          }
        }
      if (ninput2)
        {
        *outpix = val;
        *outwpix = 1.0;
        }
      else
        {
        *outpix = 0.0;
        *outwpix = BIG;
        }
      break;
    case COADD_MAX:
      ninput2 = 0;
      val = -BIG;
      wval = 0.0;
      for (i=ninput; i--;)
        {
        val2 = *(inpix++);
        wval2 = *(inwpix++);
        if (wval2<wthresh && val2>val)
          {
          ninput2++;
          val = val2;
          }
        }
      if (ninput2)
        {
        *outpix = val;
        *outwpix = 1.0;
        }
      else
        {
        *outpix = 0.0;
        *outwpix = BIG;
        }
      break;
    case COADD_WEIGHTED:
      ninput2 = 0;
      wval = val = 0.0;
      for (i=ninput; i--;)
        {
        val2 = *(inpix++);
        wval2 = *(inwpix++);
        if (wval2<wthresh)
          {
          ninput2++;
          val += val2/wval2;
          wval += 1.0/wval2;
          }
        }
      if (ninput2)
        {
        *outpix = val/wval;
        *outwpix = 1.0/wval;
        }
      else
        {
        *outpix = 0.0;
        *outwpix = BIG;
        }
      break;
    case COADD_CHI2:
      ninput2 = 0;
      wval = val = 0.0;
      for (i=ninput; i--;)
        {
        val2 = *(inpix++);
        wval2 = *(inwpix++);
        if (wval2<wthresh)
          {
          ninput2++;
          val += val2*val2/wval2;
          }
        }
      if (ninput2)
        {
        *outpix = sqrt(val/ninput2);
        *outwpix = 1.0;
        }
      else
        {
        *outpix = 0.0;
        *outwpix = BIG;
        }
      break;
    case COADD_SUM:
      ninput2 = 0;
      wval = val = 0.0;
      for (i=ninput; i--;)
        {
        val2 = *(inpix++);
        wval2 = *(inwpix++);
        if (wval2<wthresh)
          {
          ninput2++;
          val += val2;
          wval += wval2;
          }
        }
      if (ninput2)
        {
        *outpix = val;
        *outwpix = wval;
        }
      else
        {
        *outpix = 0.0;
        *outwpix = BIG;
        }
      break;
    default:
      error(EXIT_FAILURE, "*Internal Error*: Unknown Combine option in ",
			"coadd_pix()");
    }

  return RETURN_OK;
  }


/******* fast_median **********************************************************
PROTO	PIXTYPE fast_median(PIXTYPE *arr, int n)
PURPOSE	Fast median from an input array, optimized version based on the
	select() routine (Numerical Recipes, 2nd ed. Section 8.5 and
	http://www.eso.org/~ndevilla/median/). If n is even, then the result
	is the average of the 2 "central" values.
INPUT	Input pixel array ptr,
	number of input pixels,
OUTPUT	Value of the median.
NOTES	n must be >0. Warning: changes the order of data (but does not sort
	them)!
AUTHOR	E. Bertin (IAP)
VERSION	19/08/2000
 ***/
#define MEDIAN_SWAP(a,b) { register PIXTYPE t=(a);(a)=(b);(b)=t; }

PIXTYPE fast_median(PIXTYPE *arr, int n) 
  {
   PIXTYPE	*alow, *ahigh, *amedian, *amiddle, *all, *ahh,
		val, valmax, valmax2;
   int		i;

  alow = arr;
  ahigh = arr + n - 1;
  amedian = arr + n/2;
  while (ahigh > (all=alow + 1))
    {
/*-- Find median of low, middle and high items; swap into position low */
    amiddle = alow + (ahigh-alow)/2;
    if (*amiddle > *ahigh)
      MEDIAN_SWAP(*amiddle, *ahigh);
    if (*alow > *ahigh)
      MEDIAN_SWAP(*alow, *ahigh);
    if (*amiddle > *alow)
      MEDIAN_SWAP(*amiddle, *alow);

/*-- Swap low item (now in position middle) into position (low+1) */
    MEDIAN_SWAP(*amiddle, *all);

/*-- Nibble from each end towards middle, swapping items when stuck */
    ahh = ahigh;
    for (;;)
      {
      while (*alow > *(++all));
      while (*(--ahh) > *alow);

      if (ahh < all)
        break;

      MEDIAN_SWAP(*all, *ahh);
      }

/*-- Swap middle item (in position low) back into correct position */
    MEDIAN_SWAP(*alow, *ahh) ;

/*-- Re-set active partition */
    if (ahh <= amedian)
      alow = all;
    if (ahh >= amedian)
      ahigh = ahh - 1;
    }

/* One or two elements left */
  if (*alow > *ahigh)
    MEDIAN_SWAP(*alow, *ahigh);

  if (n&1)
/*-- Odd case */
    return *amedian;
  else
    {
/*-- Even case */
    valmax2 = *amedian;
    valmax = -BIG;
    alow = arr;
    for (i=n;i--;)
      if ((val=*(alow++)) > valmax && val<valmax2)
        valmax = val;
    return (*amedian+valmax)/2.0;
    }

  }

#undef MEDIAN_SWAP


/******************************** hmedian ***********************************/
/*
Median using Heapsort algorithm (for float arrays) (based on Num.Rec algo.).
Warning: changes the order of data!
*/
PIXTYPE   hmedian(PIXTYPE *ra, int n)

  {
   int          l, j, ir, i;
   float        rra;


  if (n<2)
    return *ra;
  ra--;
  for (l = ((ir=n)>>1)+1;;)
    {
    if (l>1)
      rra = ra[--l];
    else
      {
      rra = ra[ir];
      ra[ir] = ra[1];
      if (--ir == 1)
        {
        ra[1] = rra;
        return n&1? ra[n/2+1] : (ra[n/2]+ra[n/2+1])/2.0;
        }
      }
    for (j = (i=l)<<1; j <= ir;)
      {
      if (j < ir && ra[j] < ra[j+1])
        ++j;
      if (rra < ra[j])
        {
        ra[i] = ra[j];
        j += (i=j);
        }
      else
        j = ir + 1;
      }
    ra[i] = rra;
    }

/* (the 'return' is inside the loop!!) */
  }

