
/*----------------------------------------------------------------------------
   
   File name    :   imstats.c
   Author       :   Thomas Erben
   Created on   :   31.01.2005
   Description  :   estimates statistics for a set of FITS images

 ---------------------------------------------------------------------------*/

/*
	$Id: imstats.c,v 1.16 2007-09-18 13:43:07 thomas Exp $
	$Author: thomas $
	$Date: 2007-09-18 13:43:07 $
	$Revision: 1.16 $
*/

/*----------------------------------------------------------------------------
                                History Coments
 ---------------------------------------------------------------------------*/

/*
  31.01.2005:
  project started.

  01.02.2005:
  - I corrected an error in the reading
    of command line arguments (the index
    pointing to the argument to read was
    no longer correct when the filenames
    were not the last argument).
  - I corrected the form of CVS keywords for
    'keyword substitution'.

  07.02.2005:
  - I added memory tracing capabilities by the inclusion
    of 'xmemory'.

  - I cleanly free memory at the end of the program.

  25.02.2005:
  I corrected a bug in the error message when a FITS file
  could not be opened.

  20.03.2005:
  I corrected a bug showing up when several images with different
  sizes were processed but no limits on the stats region were
  specified.

  12.09.2005:
  I corrected a bug in the definition of the
  imagenames array. The two dimensions giving
  the maximum number of images and the maximum
  number of string elements per entry were in
  the wrong order.

  28.03.2006:
  I included the lower and upper quartiles in the
  statistics output.

  25.04.2006:
  I corrected a bug in the 'usage' definition.

  27.04.2006:
  I corrected a bug in the scanning of input filenames.
  A core dump could occur if a filename had more than MAXCHAR
  characters.

  17.05.2006:
  I deleted a never used parameter of the 'usage' function.

  22.03.2007:
  I included the output of number of 'good' pixels in absolute
  number ans as percent of the sample that is considered for 
  statistics. Good pixels are those which liw within the user 
  defined rejection criteria. It was included to be able to 
  test how many percent of a FITS image have saturated pixels.

  10.09.2007:
  I corrected a bug when converting a command line argument to
  'int' instead of to 'float'.

  18.09.2007:
  The number of decimal points output in the results can be given
  as command line option.
 */

/*----------------------------------------------------------------------------
                                Includes and Defines
 ---------------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>

#include "stats_stuff.h"
#include "error.h"
#include "arrays.h"
#include "fits.h"
#include "global_vers.h"
#include "t_iso8601.h"
#include "xmemory.h"

#ifndef MAXCHAR
#define MAXCHAR 255
#endif 

/* uncomment the following line to obtain a
   status on allocated/freed memory at the end
   of the program
*/

/* #define MEMORY_DEBUG */

#define __USAGE__ "\
USAGE:\n\
   imstats images [options]\n\n\
OPTIONS:\n\
-h                            print this and leave\n\
-o                            output file\n\
     (defaulted to stdout)\n\
-s   xmin xmax ymin ymax      define statistic region\n\
     (defaulted to entire frame)\n\
-t   lo_rej hi_rejj           define low/high threshholds\n\
                              for pixels to be considered\n\
     (defaulted to -60000.0 60000.0)\n\
-d                            number of decimal points in\n\
                              output numbers\n\
     (defaulted to 2)\n\n\
DESCRIPTION:\n\
The program estimates statistics from the given FITS images.\n\
It provides mode, lower quartile, median, upper quartile, mean and\n\
sigma in a tabulated form. Lower/upper quartile give the points where\n\
the integrated propability function of the sample points reaches 1/4\n\
and 3/4 respectively.\n\n\
EXAMPLE:\n\
   imstats *.fits -s 500 1500 1500 2500 -t -50000.0 50000.0 -o stat.dat\n\n\
This estimates from all FITS files in the current directory\n\
statistics for the given area within the specified pixel value\n\
region and prints the result to the file 'stat.dat'.\n\n\
AUTHOR:\n\
   Thomas Erben (terben@astro.uni-bonn.de)\n\n\
The code is heavily based on Nick Kaiser's 'stats' routine within\n\
imcat.\n"


#define MAXIMAGE 10000 /* the maximum number of input images allowed */

#define __PROGRAM__ "imstats: estimate statistics from a set of images"
#define __VERS__ "1.2.0"


static const char notokstr[] = {" \t=,;\n\r\""};
static void usage() ;

/*----------------------------------------------------------------------------
                            Main code
 ---------------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
   int		a, i;
   int          N1old = 0, N2old = 0, N1, N2;
   int          nimages = 0;
   float	**f = NULL;           /* holds pixel data of input images */
   fstatsrec 	srec;
   fitsheader	*fits;
   FILE		*ipf, *opf = stdout;
   float        lo_rej = -60000.0;    /* low pixel reject. threshhold for
                                         statistics */
   float        hi_rej = 60000.0;     /* high pixel reject. threshhold for
                                         statistics */
   int          ndecimals = 2;        /* number of decimal points in 
                                         output numbers */
   int          xmin = 0, xmax = 0;   /* x-range for statistic estimation */
   int          ymin = 0, ymax = 0;   /* y-range for statistic estimation */
   int          rangespec = 0;        /* was a x and y-range specified ? */
   char         imagenames[MAXIMAGE][MAXCHAR];
   char         outfilename[MAXCHAR]="";
   char         tmpstring[MAXCHAR];
   char         formatstring[MAXCHAR];
   char         *str;

   /* read comand line arguments */
   if(argc < 2)
   {
       usage(argv[0]);
   }

   for (a=1; a<argc; a++)
   {
     if (argv[a][0] == '-')
     {
         switch((int)tolower((int)argv[a][1]))
         {
 /*-------- Config filename */
           case 'h':    usage(argv[0]); 
  		        break;
           case 'd':	if(a<(argc-1))
	                {
			    ndecimals = atoi(argv[++a]);
			}
	                else
		        {
			    usage(argv[0]);  
			}
  			break;
           case 'o':	if(a<(argc-1))
	                {
			    strncpy(outfilename, argv[++a], MAXCHAR-1);
			}
	                else
		        {
			    usage(argv[0]);  
			}
  			break;
           case 's':	if(a<(argc-4))
	                {
			    xmin = atoi(argv[++a]);
			    xmax = atoi(argv[++a]);
			    ymin = atoi(argv[++a]);
			    ymax = atoi(argv[++a]);
			    rangespec = 1;
			}
	                else
		        {
			    usage(argv[0]);  
			}
  			break;
           case 't':	if(a<(argc-2))
	                {
			    lo_rej = atof(argv[++a]);
			    hi_rej = atof(argv[++a]);
			}
	                else
		        {
			    usage(argv[0]);  
			}
  			break;
           default:	usage(argv[0]);
		        break;
         }
     }
     else
     {
 /*---- The input image filename(s) */
       for(; (a<argc) && (*argv[a]!='-'); a++)
       {
         for (str=NULL;(str=strtok(str?NULL:argv[a], notokstr)); nimages++)
	 {
           if (nimages<MAXIMAGE)
	   {
	       strncpy(imagenames[nimages], str, MAXCHAR-1);  
	   }
           else
	   {
             error_exit("*Error*: Too many input images (more than 10000)");
	   }
	 }
       }
       a--;
     }
   }

   if(strcmp(outfilename, "") != 0)
   {
       if((opf = fopen(outfilename, "w")) == NULL)
       {
	   snprintf(tmpstring, MAXCHAR-1, "could not open %s for output: aborting\n", 
                    outfilename);
	   error_exit(tmpstring);
       }
   }

   if(ndecimals < 0)
   {
     ndecimals = 2;
   }

   /* write start message */
   fprintf(opf, "# imstats: %s\n", __PIPEVERS__);
   fprintf(opf, "# imstats: called at %s\n", get_datetime_iso8601());
   fprintf(opf, "# imstats: estimating statistics in\n");
   if(xmax == 0 || ymax == 0)
   {
     fprintf(opf, "# the whole frames within %f < pixvalue < %f\n", 
             lo_rej, hi_rej);
   }
   else
   {
     fprintf(opf, "# %d < x < %d, %d < y < %d within %f < pixvalue < %f\n", 
             xmin, xmax, ymin, ymax, lo_rej, hi_rej);
   }
   fprintf(opf, "#\n");
   fprintf(opf, "# imstats: filename\tmode\tlquartile\tmedian\tuquartile\tmean\tsigma\tgoodpix\tgoodpix/percent\n");

   for(i=0; i < nimages; i++)
   {
       if((ipf=fopen(imagenames[i], "r")) == NULL)
       {
	   snprintf(tmpstring, MAXCHAR-1, "could not open file %s: aborting!!\n", imagenames[i]);
	   error_exit(tmpstring);
       }   

       /* read in FITS image header */
       fits = readfitsheader(ipf);
       N1 = fits->n[0];
       N2 = fits->n[1];

       if((N1 != N1old) || (N2 != N2old))
       {
	   freeFloatArray(f);
	   allocFloatArray(&f, N1, N2);
	   N1old = N1;
	   N2old = N2;
       }

       /* read in FITS image pixel data */
       readfitsplane((void **)f, fits);       

       /* if limits have not been given we take the whole frame */
       if(0 == rangespec)
       {
	   xmin = ymin = 0;
	   xmax = N1;
	   ymax = N2;
       }

       fdo_stats(f, N1, N2, xmin, xmax, ymin, ymax, 
                 lo_rej, hi_rej, &srec);
       
       snprintf(formatstring, MAXCHAR-1, 
                "%%s\t%%6.%df\t%%6.%df\t%%6.%df\t%%6.%df\t%%6.%df\t%%6.%df\t%%ld\t%%6.%df\n", 
                ndecimals, ndecimals, ndecimals, ndecimals, ndecimals, ndecimals, ndecimals);

       fprintf(opf, formatstring,
	       imagenames[i],
	       srec.fmode, srec.flowerquartile, srec.fmedian, 
               srec.fupperquartile, srec.fmean,
               srec.sigma, srec.goodpix, 
               (float)srec.goodpix*100./((xmax-xmin)*(ymax-ymin))); 

       delfitsheader(fits);

       if(0 != fclose(ipf))
       {
	   snprintf(tmpstring, MAXCHAR-1, "could not close %s: aborting", imagenames[i]);
	   error_exit(tmpstring);
       }
   }

   freeFloatArray(f);

   if(strcmp(outfilename, "") != 0)
   {
       fclose(opf);
   }

#ifdef MEMORY_DEBUG
   xmemory_status() ;
#endif    

   return 0;
}

/*
 * This function only gives the usage for the program
 */

static void usage()
{
    printf("%s (V%s)\n\n%s", __PROGRAM__, __VERS__, __USAGE__) ;
    exit(1) ;
} 




